"""
ykval.db
~~~~~~~~

Database Handler
"""

import logging

from .config import settings

logger = logging.getLogger(__name__)

class DBHandler(object):
    """ Database handler wrapper """
    def __init__(self, db):
        self.settings = settings['DATABASES'][db]
        # This is not nice, but works well.
        # We need to catch OperationalError in the _execute functon, so
        # the db driver has to be available
        if self.settings.get('ENGINE', 'mysql') == 'mysql':
            import MySQLdb as dbdriver
        elif self.settings['ENGINE'] == 'postgres':
            import psycopg2 as dbdriver
        elif self.settings['ENGINE'] == 'sqlite':
            import sqlite3 as dbdriver
        else:
            raise ValueError('Invalid Database configuration')
        self.dbdriver = dbdriver
        self._connect()

    def _connect(self):
        """ Connect to the database """
        if self.settings.get('ENGINE', 'mysql') == 'mysql':
            self._db = self.dbdriver.connect(self.settings['HOST'],
                                             self.settings['USER'],
                                             self.settings['PASSWORD'],
                                             self.settings['NAME'],
                                             self.settings.get('PORT', 3306))
        elif self.settings['ENGINE'] == 'postgres':
            self._db = self.dbdriver.connect(database=self.settings['NAME'],
                                             user=self.settings['USER'],
                                             password=self.settings['PASSWORD'],
                                             host=self.settings['HOST'])
        elif self.settings['ENGINE'] == 'sqlite':
            self._db = self.dbdriver.connect(self.settings['NAME'])
        self.cursor = self._db.cursor()

    def _execute(self, query, params=None, retry=False):
        """ Abstract the cursor execute function to handle sqlite syntax """
        if self.settings.get('ENGINE') == 'sqlite':
            if '%s' in query and params:
                query = query.replace('%s', '?')
            elif '%(' in query and params:
                query = query.replace('%(', ':').replace(')s', '')
        logger.debug('QUERY: %s PARAMS: %s', query, params)
        try:
            rowcount = self.cursor.execute(query, params)
            self._db.commit()
        except (AttributeError, self.dbdriver.OperationalError) as err:
            if not retry:
                logger.debug('Database reconnect due to error: %s', err)
                self._connect()
                return self._execute(query, params, retry=True)
            else:
                raise
        except Exception as err:
            logger.exception('Database error: %s', err)
            raise

    def _dictfetchall(self):
        """ Wrapper to return DB results in dict format """
        return [dict(zip([col[0] for col in self.cursor.description], row)) \
                for row in self.cursor.fetchall()]

    def _dictfetchone(self):
        """ Wrapper to return DB results in dict format """
        data = self._dictfetchall()
        if data:
            return data[0]
        return {}



"""
ykval.crypt
~~~~~~~~~~~

Utilities to load encrypted private keys
"""

import base64
import binascii
import glob
import hashlib
import logging
import os
import string
import sys
import termios
import tty

from Crypto.Cipher import DES, DES3, AES, PKCS1_OAEP
from Crypto.Hash import MD5
from Crypto.Protocol.KDF import PBKDF1
from Crypto.PublicKey import RSA

# Older versions of PyCrypto don't skip pad/unpad, we only care about the
# PKCS#7 variants here.
try:
    from Crypto.Util.Padding import pad, unpad
except ImportError:
    def pad(data, block_size):
        """Pad a block of data to align with block_size using PKCS#7 padding."""
        return data + chr(len(data)) * ((block_size - len(data)) % block_size)

    def unpad(data, block_size):
        """Remove block padding from PKCS#7 padding."""
        if len(data) % block_size:
            raise ValueError('Input data is not padded')

        if isinstance(data[-1], int):
            pad_len = data[-1]
        else:
            pad_len = ord(data[-1])

        if pad_len < 1 or pad_len > min(block_size, len(data)):
            raise ValueError('Incorrect padding')

        padding = chr(pad_len) * pad_len
        if isinstance(data, bytes):
            padding = padding.encode('ascii')

        if data[-pad_len:] != padding:
            raise ValueError('PKCS#7 padding incorrect')

        else:
            return data[:-pad_len]


logger = logging.getLogger(__name__)


EPILOG = '''
commands:
  secret                generate a new shared secret
  test-secret           reconstruct a shared secret from shares
  test <keydir>         encrypt & decrypt <rounds> number of keys

in order to start using the encrypter/decrypter, you have to set up a key
directory first:

  ~$ mkdir keys
  keys$ # for --krn=1:
  keys$ openssl genrsa -out 1.private
  ...
  keys$ # for --krn=2:
  keys$ openssl genrsa -out 2.private -aes128
  ...
  keys$ openssl rsa -in 1.private -pubout > 1.public
  keys$ openssl rsa -in 2.private -pubout > 2.public

now you can split the secret used to protect 2.private using the secret
command:

  ~$ python yubistack/crypt.py secret

'''


class SHA2(object):
    """Wrapper for the SHA-256 hashing algorithm."""
    digest_size = 32

    def __init__(self, *args):
        if args and args[0] is object():
            self._hash = args[0].copy()
        else:
            self._hash = hashlib.new('sha256')

    def copy(self):
        """Make a (mutable) copy of the hash."""
        return SHA2(self._hash.copy())

    def digest(self, *args):
        """Calculate the hash digest."""
        return self._hash.digest(*args)

    def hexdigest(self, *args):
        """Calculate the hash hexdigest."""
        return self._hash.hexdigest(*args)

    def update(self, *args):
        """Update the hash."""
        return self._hash.update(*args)


SHA2.new = SHA2


class Crypter(object):
    """Crypter base class for doing encryption and decryption."""

    mask = '*.*'
    shares = 2

    def __init__(self, keydir):
        # Keydir path
        self.keydir = os.path.abspath(keydir)
        # Our key rotation number
        self.krn = 0
        self.key = {}

        # Load keyfile with highest krn from disk
        keyfiles = os.path.join(self.keydir, self.mask)
        for path in sorted(glob.glob(keyfiles)):
            part = os.path.basename(path).split('.')
            if not part[-2].isdigit():
                continue

            krn = int(part[-2])
            if krn > self.krn:
                self.krn = krn

            self.key[krn] = self.load_keyfile(path)

        if self.krn == 0:
            raise ValueError('No key was loaded')

        logging.info('Crypter krn=%d', self.krn)

    def cipher(self, krn=None):
        """Generate a cipher for the given key rotation number."""
        return PKCS1_OAEP.new(
            self.key[krn or self.krn],  # key
            SHA2,                       # digest
        )

    def decrypt(self, ciphertext, krn=None):
        """Decrypt a ciphertext."""
        return self.cipher(krn).decrypt(ciphertext)

    def encrypt(self, plaintext, krn=None):
        """Encrypt a plaintext."""
        return self.cipher(krn).encrypt(plaintext)

    def load_keyfile(self, keyfile):
        """Load a key file."""
        raise NotImplementedError('Implement me in subclass')


class PublicKey(Crypter):
    """Public key that can encrypt."""

    mask = '*.public'

    def decrypt(self, ciphertext, krn=None):
        raise NotImplementedError("Public key can only encrypt")

    def load_keyfile(self, keyfile):
        with open(keyfile) as handle:
            return RSA.importKey(handle.read())


class PrivateKey(Crypter):
    """Private key that can both encrypt and decrypt.

    The private key may be encrypted using 3DES-CBC or AES-128-CBC, if so, it
    is assumed that the decryption secret is split using Shamir's Shared Secret
    algorithm.
    """

    mask = '*.private'

    def load_keyfile(self, keyfile):
        with open(keyfile) as handle:
            keydata = handle.read()
            try:
                return RSA.importKey(keydata)
            except ValueError:
                return RSA.importKey(self.load_encrypted_keydata(keydata))

    def load_encrypted_keydata(self, keydata):
        """Load a key from encrypted keydata (in PEM format)."""
        lines = keydata.strip().replace(' ', '').splitlines()
        if not lines[1].startswith('Proc-Type:4,ENCRYPTED'):
            raise TypeError('Unsupported encryption')

        dek = lines[2].split(':')
        if len(dek) != 2 or dek[0] != 'DEK-Info':
            raise ValueError('PEM encryption method not supported')

        algo, salt = dek[1].split(',')
        salt = binascii.unhexlify(salt)

        if algo == 'DES-CBC':
            key = PBKDF1(self.passphrase, salt, 8, 1, MD5)
            obj = DES.new(key, DES.MODE_CBC, salt)

        elif algo == 'DES-EDE3-CBC':
            key = PBKDF1(self.passphrase, salt, 16, 1, MD5)
            key += PBKDF1(key + self.passphrase, salt, 8, 1, MD5)
            obj = DES3.new(key, DES3.MODE_CBC, salt)

        elif algo == 'AES-128-CBC':
            key = PBKDF1(self.passphrase, salt[:8], 16, 1, MD5)
            obj = AES.new(key, AES.MODE_CBC, salt)

        else:
            raise TypeError('%s: cipher not supported' % (algo,))

        lines = lines[3:-1]
        data = base64.b64decode(''.join(lines))
        return unpad(obj.decrypt(data), obj.block_size)

    @property
    def passphrase(self):
        if not hasattr(self, '_passphrase'):
            try:
                return os.environ['YKKSM_PASSPHRASE']
            except KeyError:
                self._passphrase = self._getpass_cooked('Private key passphrase: ')
                os.environ['YKKSM_PASSPHRASE'] = self._passphrase
        return self._passphrase

    def _getpass_cooked(self, prompt):
        sys.stdout.write(prompt + '\r\n')
        phrase = []

        fd = sys.stdin.fileno()
        old = termios.tcgetattr(fd)
        flg = termios.TCSAFLUSH
        if hasattr(termios, 'TCSASOFT'):
            flg |= termios.TCSASOFT

        try:
            tty.setraw(fd)
            sys.stdout.flush()
            while True:
                key = sys.stdin.read(1)
                if key in '\r\n': # return
                    sys.stdout.write('\n')
                    sys.stdout.flush()
                    return ''.join(phrase)
                elif key == chr(0x7f): # backspace
                    sys.stdout.write('\r' + ' ' * len(phrase))
                    if phrase:
                        phrase = phrase[:-1]
                    sys.stdout.write('\r' + '*' * len(phrase))
                    sys.stdout.flush()
                elif key == chr(0x03): # ^C
                    raise KeyboardInterrupt()
                else:
                    if key in string.printable:
                        phrase.append(key)
                        sys.stdout.write('*')
                        sys.stdout.flush()
        finally:
            termios.tcsetattr(fd, flg, old)


def run():
    """Main entry point."""
    import argparse

    logging.basicConfig(
        format='%(asctime)s %(levelname)s %(module)s %(funcName)s: %(message)s',
        level=logging.DEBUG,
    )

    parser = argparse.ArgumentParser(
        epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('-r', '--rounds', type=int, default=100)
    parser.add_argument('-k', '--krn', type=int, default=None)
    parser.add_argument('--pub', default=False, action='store_true')
    parser.add_argument('-d', '--keydir', default='/etc/yubico/ksm-keys')
    parser.add_argument('command')
    args = parser.parse_args()

    if args.command == 'encrypt':
        crypter = PublicKey(args.keydir)
        while True:
            plaintext = input('plain text: ')
            if not plaintext:
                break

            ciphertext = crypter.encrypt(plaintext.encode('utf-8'), krn=args.krn)
            print(base64.b64encode(ciphertext).decode('ascii'))

    elif args.command == 'test':
        if args.pub:
            crypter = PublicKey(args.keydir)
        else:
            crypter = PrivateKey(args.keydir)

        logger.info('running %d test rounds', args.rounds)
        for r in range(args.rounds):
            p = os.urandom(16).encode('hex')
            print('encrypt', p, len(p))
            c = crypter.encrypt(p, krn=args.krn)
            print(base64.b64encode(c))
            if not args.pub:
                assert crypter.decrypt(c, krn=args.krn) == p, 'round %d failed' % (r,)
        logger.info('all good')

    else:
        print('No such command {!r}'.format(args.command))


if __name__ == '__main__':
    sys.exit(run())



"""
yubistack.ykauth
~~~~~~~~~~~~~~~~

Python Yubikey Stack - Authentication module
"""

import base64
import logging
import urllib

from passlib.context import CryptContext
import requests

from .config import (
    settings,
    TOKEN_LEN,
)
from .db import DBHandler
from .exceptions import YKAuthError
from .utils import (
    sign,
    generate_nonce,
)

logger = logging.getLogger(__name__)


class DBH(DBHandler):
    """
    Extending the generic DBHandler class with the required queries
    """
    def get_user(self, username):
        """
        Read user information for Yubiauth
        """
        query = """SELECT users.attribute_association_id AS users_attribute_association_id,
                          users.id AS users_id, users.name AS users_name,
                          users.auth AS users_auth
                     FROM users
                    WHERE users.name = %s"""
        self._execute(query, (username,))
        return self._dictfetchone()

    def get_token(self, user_id, token_id):
        """
        Read user attribute information for Yubiauth
        """
        query = """SELECT yubikeys.attribute_association_id AS yubikeys_attribute_association_id,
                          yubikeys.id AS yubikeys_id,
                          yubikeys.prefix AS yubikeys_prefix,
                          yubikeys.enabled AS yubikeys_enabled
                     FROM yubikeys
               INNER JOIN user_yubikeys
                       ON user_yubikeys.yubikey_id = yubikeys.id
                    WHERE user_yubikeys.user_id = %s
                      AND yubikeys.prefix = %s"""
        self._execute(query, (user_id, token_id))
        return self._dictfetchone()


class VerificationClient(object):
    """ Verification Client """
    def __init__(self, urls, client_id=None, apikey=None):
        self.urls = urls
        self.client_id = client_id
        self.apikey = base64.b64decode(apikey)

    def generate_query(self, otp, nonce, timestamp=False, timeout=None,
                       sync_level=None):
        """ Generate query """
        data = [('id', self.client_id),
                ('otp', otp),
                ('nonce', nonce)]
        if timestamp:
            data.append(('timestamp', '1'))

        if sync_level is not None:
            data.append(('sl', sync_level))

        if timeout:
            data.append(('timeout', timeout))

        query_string = urllib.parse.urlencode(data)
        if self.apikey:
            signature = sign(dict(data), self.apikey)
            query_string += '&h=%s' % (signature.replace('+', '%B'))
        return query_string

    def verify(self, otp, timestamp=False, sl=None, timeout=None,
               return_response=False):
        """ Make a HTTP call to the Yubikey Verification servers """
        nonce = generate_nonce()
        query = self.generate_query(otp, nonce, timestamp=timestamp,
                                    timeout=timeout, sync_level=sl)
        req = requests.get(self.urls[0] + '?' + query)
        print(req.text)


class Client(object):
    """ Authentication Client """
    def __init__(self):
        self.db = DBH(db='yubiauth')
        self.pwd_context = CryptContext(**settings['CRYPT_CONTEXT'])
        if settings['USE_NATIVE_YKVAL']:
            # Native verify
            from .ykval import Validator
            self.ykval_client = Validator()
        else:
            # Using yubico_client to verify against remote server
            from yubico_client import Yubico
            self.ykval_client = Yubico(settings['YKVAL_CLIENT_ID'],
                                       settings['YKVAL_CLIENT_SECRET'],
                                       api_urls=settings['YKVAL_SERVERS'])

    def _get_user_info(self, username):
        """
        Get user from DB

        Args:
            username

        Returns:
            dictionary of user data

        Raises:
            AuthFail if user does not exist
        """
        user = self.db.get_user(username)
        if not user:
            raise YKAuthError('UNKNOWN_USER')
        logger.debug('Found user: %s', user)
        return user

    def _check_token(self, user, token_id):
        """
        Check Token association with user

        Args:
            user: User data dict as recieved from _get_user_info()
            token_id: Token prefix (aka. publicname)

        Returns:
            None

        Raises:
            AuthFail if token is not associated with the user
            AithFail if token is disabled
        """
        token = self.db.get_token(user['users_id'], token_id)
        if not token:
            logger.error('Token %s is not associated with %s', token_id, user['users_name'])
            raise YKAuthError('INVALID_TOKEN')
        logger.debug('Found token: %s', token)
        if not token.get('yubikeys_enabled'):
            logger.error('Token %s is disabled for %s', token_id, user['users_name'])
            raise YKAuthError('DISABLED_TOKEN')

    def _validate_password(self, user, password):
        """
        Validate password against the hash in SQL
        """
        valid, new_hash = self.pwd_context.verify_and_update(str(password), user['users_auth'])
        if not valid:
            logger.error('%(users_name)s: Invalid password', user)
            raise YKAuthError('BAD_PASSWORD')
        if new_hash:
            # TODO: update user's hash with new_hash
            logger.warning("User %(users_name)s's hash needs update", user)
        return True

    def _validate_otp(self, otp):
        """
        Use Yubico client to validate OTP
        """
        try:
            if self.ykval_client.verify(otp):
                return True
            return False
        except Exception as err:
            logger.error('OTP Validation failed: %r', err)
            return False

    def authenticate(self, username, password, otp):
        """
        Yubistack user authentication

        Args:
            username: Username of the user
            password: Password/PIN of the user
            otp: Yubikey one time password

        Returns:
            dict of authentication data

        Authentication process:
            1. Check if token is enabled
            2. Check if token is associated with the user & enabled
            3. Validate users password
            4. Validate OTP (YKVal)
        """
        token_id = otp[:-TOKEN_LEN]
        # STEP 1: Check if token is enabled
        user = self._get_user_info(username)
        # STEP 2: Check if token is associated with the user & enabled
        self._check_token(user, token_id)
        # STEP 3: Validate users password
        self._validate_password(user, password)
        # STEP 4: Validate OTP
        self.ykval_client.verify(otp)
        return True



"""
yubistack.ykval
~~~~~~~~~~~~~~~

Python Yubikey Stack - Key validation module
"""

import base64
from datetime import datetime
import logging
try:
    import queue
except ImportError:
    import Queue as queue
import re
import threading
import time

import requests

from .config import (
    settings,
    TS_SEC,
    TS_REL_TOLERANCE,
    TS_ABS_TOLERANCE,
    TOKEN_LEN,
    OTP_MAX_LEN,
)
from .db import DBHandler
from .exceptions import (
    YKValError,
    YKSyncError,
)
from .utils import (
    generate_nonce,
    counters_eq,
    counters_gt,
    counters_gte,
    parse_sync_response,
)

logger = logging.getLogger(__name__)

class DBH(DBHandler):
    """
    Extending the generic DBHandler class with the required queries
    """
    def get_client_data(self, client_id):
        """ Lookup client based on the ID """
        query = """SELECT id,
                          secret
                     FROM clients
                    WHERE active = 1
                      AND id = %s"""
        self._execute(query, (client_id,))
        return self._dictfetchone()

    def get_local_params(self, yk_publicname):
        """ Get yubikey parameters from DB """
        query = """SELECT active,
                          modified,
                          yk_publicname,
                          yk_counter,
                          yk_use,
                          yk_low,
                          yk_high,
                          nonce
                     FROM yubikeys
                    WHERE yk_publicname = %s"""
        self._execute(query, (yk_publicname,))
        local_params = self._dictfetchone()
        if not local_params:
            local_params = {
                'active': '1',
                'modified': -1,
                'yk_publicname': yk_publicname,
                'yk_counter': -1,
                'yk_use': -1,
                'yk_low': -1,
                'yk_high': -1,
                'nonce': '0000000000000000',
                'created': int(time.time())
            }
            # Key was missing in DB, adding it
            self.add_new_identity(local_params)
            logger.warning('Discovered new identity %s', yk_publicname)
        logger.debug('Auth data: %s', local_params)
        return local_params

    def add_new_identity(self, identity):
        """ Create new key identity """
        query = """INSERT INTO yubikeys (
                       active,
                       created,
                       modified,
                       yk_publicname,
                       yk_counter,
                       yk_use,
                       yk_low,
                       yk_high,
                       nonce
                ) VALUES (
                       %(active)s,
                       %(created)s,
                       %(modified)s,
                       %(yk_publicname)s,
                       %(yk_counter)s,
                       %(yk_use)s,
                       %(yk_low)s,
                       %(yk_high)s,
                       %(nonce)s
                )"""
        self._execute(query, identity)

    def get_queue(self, modified, server_nonce):
        """
        Read all elements from queue
        """
        query = """SELECT server,
                          otp,
                          modified,
                          info
                     FROM queue
                    WHERE modified=%s
                      AND server_nonce = %s"""
        self._execute(query, (modified, server_nonce))
        return self._dictfetchall()

    def remove_from_queue(self, server, modified, server_nonce):
        """
        Remove a single element from queue
        """
        query = """DELETE FROM queue
                         WHERE server = %s
                           AND modified = %s
                           AND server_nonce = %s"""
        self._execute(query, (server, modified, server_nonce))

    def null_queue(self, server_nonce):
        """
        NULL queued_time for remaining entries in queue, to allow
        daemon to take care of them as soon as possible.
        """
        query = """UPDATE queue
                      SET queued = NULL
                    WHERE server_nonce = %s"""
        self._execute(query, (server_nonce,))

    def update_db_counters(self, params):
        """ Update table with new counter values """
        query = """UPDATE yubikeys
                      SET modified = %(modified)s,
                          yk_counter = %(yk_counter)s,
                          yk_use = %(yk_use)s,
                          yk_low = %(yk_low)s,
                          yk_high = %(yk_high)s,
                          nonce = %(nonce)s
                    WHERE yk_publicname = %(yk_publicname)s
                      AND (yk_counter < %(yk_counter)s
                       OR (yk_counter = %(yk_counter)s
                      AND yk_use < %(yk_use)s))"""
        self._execute(query, params)

    def enqueue(self, otp_params, local_params, server, server_nonce):
        """
        Insert new params into database queue table
        """
        info = 'yk_publicname=%(yk_publicname)s&yk_counter=%(yk_counter)s' % otp_params
        info += '&yk_use=%(yk_use)s&yk_high=%(yk_high)s&yk_low=%(yk_low)s' % otp_params
        info += '&nonce=%(nonce)s' % otp_params
        info += ',&local_counter=%(yk_counter)s&local_use=%(yk_use)s' % local_params
        query = """INSERT INTO queue (
                        queued,
                        modified,
                        otp,
                        server,
                        server_nonce,
                        info
                    ) VALUES (%s, %s, %s, %s, %s, %s)"""
        self._execute(query, (int(time.time()), otp_params['modified'],
                              otp_params['otp'], server, server_nonce, info))

    def get_keys(self, yk_publicname):
        """ Get all keys from DB """
        query = """SELECT yk_publicname
                     FROM yubikeys
                    WHERE active = 1"""
        params = None
        if yk_publicname != 'all':
            query += ' AND yk_publicname = %s'
            params = (yk_publicname,)
        self._execute(query, params)
        return self.cursor.fetchall()

REQUIRED_PARAMS = ['modified', 'otp', 'nonce', 'yk_publicname',
                   'yk_counter', 'yk_use', 'yk_high', 'yk_low']
class Sync(object):
    """ Sync object to handle cross synchronization requests """
    def __init__(self, db=None):
        self.db = db if db else DBH(db='ykval')
        self.sync_servers = settings['SYNC_SERVERS']

    def check_sync_input(self, sync_params):
        """ Check for all required parameters """
        for req_param in REQUIRED_PARAMS:
            if req_param not in sync_params:
                logger.error("Received request with missing '%s' parameter", req_param)
                raise YKSyncError('MISSING_PARAMETER', req_param)
            if req_param not in ('otp', 'nonce', 'yk_publicname') and not \
                (sync_params[req_param] == '-1' or isinstance(sync_params[req_param], int)):
                logger.error("Input parameter '%s' is not correct", req_param)
                raise YKSyncError('INVALID_PARAMETER', req_param)

    def check_resync_input(self, resync_params):
        """ Check input parameters """
        if 'yk' not in resync_params:
            logger.error("Received request with missing 'yk' parameter")
            raise YKSyncError('MISSING_PARAMETER', 'yk')
        if not re.match(r'^([cbdefghijklnrtuv]{0,16}|all)$', resync_params['yk']):
            logger.error("Invalid 'yk' value: %(yk)s", resync_params)
            raise YKSyncError('INVALID_PARAMETER', 'yk')

    def sync_local(self, sync_params):
        """ Synchronize """
        self.check_sync_input(sync_params)
        local_params = self.db.get_local_params(sync_params['yk_publicname'])
        self.db.update_db_counters(sync_params)
        logger.debug('Local params: %s', local_params)
        logger.debug('Sync request params: %s', sync_params)

        if counters_gte(local_params, sync_params):
            logger.warning('%(yk_publicname)s: Remote server out of sync', sync_params)

        if counters_eq(local_params, sync_params):
            if sync_params['modified'] == local_params['modified'] \
                and sync_params['nonce'] == local_params['nonce']:
                # This is not an error. When the remote server received
                # an OTP to verify, it would have sent out sync requests
                # immediately. When the required number of responses had
                # been received, the current implementation discards all
                # additional responses (to return the result to the client
                # as soon as possible). If our response sent last time was
                # discarded, we will end up here when the background
                # ykval-queue processes the sync request again.
                logger.info('%(yk_publicname)s: Sync request unnecessarily sent', sync_params)

            if (
                    sync_params['modified'] != local_params['modified'] and
                    sync_params['nonce'] == local_params['nonce']
                ):
                delta_modified = sync_params['modified'] - local_params['modified']
                if delta_modified < -1 or delta_modified > 1:
                    logger.warning('%s: We might have a replay attack. 2 events '
                                   'at different times have generated the same '
                                   'counters. Time difference is %s sec',
                                   sync_params['yk_publicname'], delta_modified)

            if sync_params['nonce'] != local_params['nonce']:
                logger.warning('%(yk_publicname)s: Remote server has received a request to '
                               'validate an already validated OTP', sync_params)

        if not local_params['active']:
            # The remote server has accepted an OTP from a YubiKey which
            # we would not. We still needed to update our counters with
            # the counters from the OTP thought.
            logger.warning('%(yk_publicname)s: Received sync-request for de-activated Yubikey',
                           sync_params)
            raise YKSyncError('DISABLED_TOKEN')
        return local_params

    def resync_local(self, resync_params):
        """ Re-synchronize """
        self.check_resync_input(resync_params)
        keys = self.db.get_keys(resync_params['yk'])
        server_nonce = generate_nonce()
        for key in keys:
            local_params = self.db.get_local_params(key)
            local_params['otp'] = 'c' * 32 # Fake an OTP
            logger.debug('Auth data: %s', local_params)
            for server in self.sync_servers:
                self.db.enqueue(local_params, local_params, server, server_nonce)
        return 'OK Initiated resync of %(yk)s' % resync_params

    def _fetch_remote(self, dqueue, server, url, timeout):
        """ Make HTTP GET call to remote server """
        try:
            req = requests.get(url, timeout=timeout)
            if req.status_code == 200:
                try:
                    resp_params = parse_sync_response(req.text)
                    dqueue.put({'server': server, 'params': resp_params})
                except ValueError as err:
                    logger.error('Failed to parse response of %s: %s', server, err)
            else:
                logger.warning('Recieved status code %s for %s', req.status_code, url)
        except Exception as err:
            logger.warning('Failed to retrieve %s: %s', url, err)

    def sync_remote(self, otp_params, local_params, server_nonce, required_answers, timeout=1):
        """ Function to synchronize values with other ykval servers """
        # Construct URLs
        responses = []
        dqueue = queue.Queue()
        for row in self.db.get_queue(otp_params['modified'], server_nonce):
            url = '%(server)s?otp=%(otp)s&modified=%(modified)s' % row
            url += '&' + row['info'].split(',')[0]
            _thread = threading.Thread(target=self._fetch_remote, args=(dqueue, row['server'],
                                                                        url, timeout))
            _thread.daemon = True
            _thread.start()
        loop_start = time.time()
        while len(responses) < required_answers and time.time() < loop_start + timeout * 1.5:
            try:
                resp = dqueue.get(timeout=0.2)
                responses.append(resp)
                # Delete entry from table
                self.db.remove_from_queue(resp['server'], otp_params['modified'], server_nonce)
            except queue.Empty:
                pass

        answers = len(responses)
        # Parse response
        valid_answers = 0
        for resp in responses:
            resp_params = resp['params']
            logger.debug('local DB contains %s', local_params)
            logger.debug('response contains %s', resp_params)
            logger.debug('OTP contains %s', otp_params)
            # Update Internal DB (conditional)
            self.db.update_db_counters(resp_params)
            # Check for Warnings
            # https://developers.yubico.com/yubikey-val/doc/ServerReplicationProtocol.html
            # NOTE: We use local_params for validationParams comparison since they are actually
            #       the same in this situation and we have them at hand.
            if counters_gt(local_params, resp_params):
                logger.warning('Remote server out of sync')
            if counters_gt(resp_params, local_params):
                logger.warning('Local server out of sync')
            if counters_eq(resp_params, local_params) \
                and resp_params['nonce'] != local_params['nonce']:
                logger.warning('Servers out of sync. Nonce differs.')
            if counters_eq(resp_params, local_params) \
                and resp_params['modified'] != local_params['modified']:
                logger.warning('Servers out of sync. Modified differs.')
            if counters_gt(resp_params, otp_params):
                logger.warning('OTP is replayed. Sync response counters higher than OTP counters.')
            elif counters_eq(resp_params, otp_params) \
                and resp_params['nonce'] != otp_params['nonce']:
                logger.warning('OTP is replayed. Sync response counters equal to OTP counters and \
nonce differs.')
            else:
                # The answer is ok since a REPLAY was not indicated
                valid_answers += 1
                if required_answers == valid_answers:
                    break

        # NULL queued_time for remaining entries in queue, to allow
        # daemon to take care of them as soon as possible.
        self.db.null_queue(server_nonce)
        return {'answers': answers, 'valid_answers': valid_answers}

class Validator(object):
    """ Yubikey OTP validator """
    def __init__(self):
        self.db = DBH(db='ykval')
        if settings['USE_NATIVE_YKKSM']:
            from .ykksm import Decryptor
            self.decryptor = Decryptor()
        else:
            self.decryptor = None
        self.sync_servers = settings['SYNC_SERVERS']
        self.default_sync_level = settings['SYNC_LEVEL']
        # Below parameters are valid from protocol version >= 2.0
        self.sync_level = None
        self.timeout = None

    def check_parameters(self, params):
        """ Perform Sanity check on parameters """
        # CLIENT ID
        if params['client_id'] and not isinstance(params['client_id'], int):
            logger.error('id provided in request must be an integer')
            raise YKValError('INVALID_PARAMETER', 'client_id')
        # OTP
        if not TOKEN_LEN <= len(params['otp']) <= OTP_MAX_LEN:
            logger.error('Incorrect OTP length: %(otp)s', params)
            raise YKValError('BAD_OTP')
        if not re.match(r'^[cbdefghijklnrtuv]+$', params['otp']):
            logger.error('Invalid OTP: %(otp)s', params)
            raise YKValError('BAD_OTP')
        # NONCE:
        # - If client_id is not provided, we're using a Native stack call
        if params['client_id'] and not params['nonce']:
            logger.error('Nonce is missing and protocol version >= 2.0')
            raise YKValError('MISSING_PARAMETER', 'nonce')
        if params['nonce'] and not re.match(r'^[A-Za-z0-9]+$', params['nonce']):
            logger.error('Nonce is provided but not correct')
            raise YKValError('INVALID_PARAMETER', 'nonce')
        if params['nonce'] and not 16 <= len(params['nonce']) <= 40:
            logger.error('Nonce too short or too long')
            raise YKValError('INVALID_PARAMETER', 'nonce')
        # TIMESTAMP
        #   NOTE: Timestamp parameter is not checked since current protocol says
        #   that 1 means request timestamp and anything else is discarded.
        # TIMEOUT
        if not isinstance(params['timeout'], int):
            logger.error('timeout is provided but not correct')
            raise YKValError('INVALID_PARAMETER', 'timeout')
        # SYNC LEVEL
        if not (isinstance(params['sync_level'], int) and 0 <= params['sync_level'] <= 100):
            logger.error('SL (sync level) is provided but not correct')
            raise YKValError('INVALID_PARAMETER', 'sync_level')

    def get_client_apikey(self, client_id):
        """
        Get Client info from DB

        Args:
            client_id: Integer ID number of the client.
                       Corresponding API key will be retrieved from DB

        Returns:
            b64decoded client secret (apikey)

        Raises:
            YKValError('NO_SUCH_CLIENT') if client doesn't exist
        """
        if not client_id:
            return ''.encode()
        client_data = self.db.get_client_data(client_id)
        logger.debug('Client data: %s', client_data)
        if not client_data:
            logger.error('Invalid client id: %s', client_id)
            raise YKValError('NO_SUCH_CLIENT')
        return base64.b64decode(client_data['secret'])

    def decode_otp(self, otp):
        """
        Call out to KSM to decrypt OTP
        """
        if self.decryptor:
            data = self.decryptor.decrypt(otp)
            return dict([(k, int(v, 16)) for k, v in data.items()])
        elif settings['YKKSM_SERVERS']:
            # TODO: Support for async req for multiple servers
            for url in settings['YKKSM_SERVERS']:
                req = requests.get(url, params={'otp': otp}, headers={'Accept': 'application/json'})
                logger.debug('YK-KSM response: %s (status_code: %s)', req.text, req.status_code)
                if req.headers['Content-Type'] == 'application/json' and req.status_code == 200:
                    return dict([(k, int(v, 16)) for k, v in req.json().items()])
                if req.text.startswith('OK'):
                    resp = {}
                    for i in req.text.split()[1:]:
                        key, val = i.split('=')
                        resp[key] = int(val, 16)
                    return resp
            raise YKValError('BAD_OTP')
        logger.error("No KSM service provided. Can't decrypt OTP.")
        raise YKValError('BACKEND_ERROR', 'No KSM service found')

    def build_otp_params(self, params, otp_info):
        """ Build OTP params """
        return {
            'modified': int(time.time()),
            'otp': params['otp'],
            'nonce': params['nonce'],
            'yk_publicname': params['otp'][:-TOKEN_LEN],
            'yk_counter': int(otp_info['counter']),
            'yk_use': int(otp_info['use']),
            'yk_high': otp_info['high'],
            'yk_low': otp_info['low'],
        }

    def validate_otp(self, otp_params, local_params):
        """ Validate OTP """
        # First check if OTP is seen with the same nonce, in such case we have an replayed request
        if counters_eq(local_params, otp_params) \
            and local_params['nonce'] == otp_params['nonce']:
            logger.warning('Replayed request')
            raise YKValError('REPLAYED_REQUEST')
        # Check the OTP counters against local DB
        if counters_gte(local_params, otp_params):
            logger.warning('Replayed OTP: Local counters higher (%s > %s)',
                           local_params, otp_params)
            raise YKValError('REPLAYED_OTP')
        # Valid OTP, update DB
        self.db.update_db_counters(otp_params)

    def replicate(self, otp_params, local_params, server_nonce):
        """ Handle sync across the cluster """
        for server in settings['SYNC_SERVERS']:
            self.db.enqueue(otp_params, local_params, server, server_nonce)

        req_answers = round(len(self.sync_servers) * float(self.sync_level) / 100.0)
        if req_answers:
            sync = Sync(self.db)
            sync_metrics = sync.sync_remote(otp_params, local_params, server_nonce,
                                            req_answers, self.timeout)
            sync_success = sync_metrics['valid_answers'] == req_answers
            sync_level_success_rate = 100.0 * sync_metrics['valid_answers'] / len(self.sync_servers)
        else:
            sync_success = True
            sync_level_success_rate = 0
            sync_metrics = {'answers': 0, 'valid_answers': 0}
        logger.info('%s: sync details: synclevel=%s server_count=%d req_answers=%d '
                    'answers=%d valid_answers=%s sl_success_rate=%.3f timeout=%s',
                    otp_params['yk_publicname'], self.sync_level, len(self.sync_servers),
                    req_answers, sync_metrics['answers'], sync_metrics['valid_answers'],
                    sync_level_success_rate, self.timeout)

        if not sync_success:
            # sync returned false, indicating that either at least 1 answer
            # marked OTP as invalid or there were not enough answers
            logger.error('ykval-verify:notice:Sync failed')
            if sync_metrics['valid_answers'] != sync_metrics['answers']:
                raise YKValError('REPLAYED_OTP')
            else:
                #self.extra['sl'] = self.sync_level_success_rate
                raise YKValError('NOT_ENOUGH_ANSWERS')

    def phishing_test(self, otp_params, local_params):
        """ Run a timing phishing test """
        # Only check token timestamps if it was not plugged out
        if otp_params['yk_counter'] != local_params['yk_counter']:
            return
        new_ts = (otp_params['yk_high'] << 16) + otp_params['yk_low']
        old_ts = (local_params['yk_high'] << 16) + local_params['yk_low']
        ts_delta = (new_ts - old_ts) * TS_SEC

        # Check real time
        last_time = local_params['modified']
        now = int(time.time())
        elapsed = now - last_time
        deviation = abs(elapsed - ts_delta)

        # Time delta server might verify multiple OTPs in a row. In such case validation server
        # doesn't have time to tick a whole second and we need to avoid division by zero.
        if elapsed:
            percent = deviation / elapsed
        else:
            percent = 1
        if deviation > TS_ABS_TOLERANCE and percent > TS_REL_TOLERANCE:
            logger.error('%s: OTP Expired:\n\t' % otp_params['otp'][:-TOKEN_LEN] +
                         'TOKEN TS OLD: %s\n\t' % datetime.utcfromtimestamp(old_ts) +
                         'TOKEN TS NEW: %s\n\t' % datetime.utcfromtimestamp(new_ts) +
                         'TOKEN TS DIFF: %s (sec)\n\t' % ts_delta +
                         'ACCESS TS OLD: %s\n\t' % datetime.utcfromtimestamp(last_time) +
                         'ACCESS TS NEW: %s\n\t' % datetime.utcfromtimestamp(now) +
                         'ACCESS TS DIFF: %s (sec)\n\t' % elapsed +
                         'DEVIATION: %s (sec) or %s%%' % (deviation, percent))
            logger.warning('OTP failed phishing test')
            raise YKValError('DELAYED_OTP')

    def verify(self, otp, client_id=None, nonce=None, timestamp=0,
               timeout=None, sync_level=None):
        """
        Yubico OTP Validation Protocol V2.0 Implementation

        Args:
            otp: The OTP from the YubiKey
            client_id: Specifies the requestor so that the end-point
                       can retrieve correct shared secret for
                       signing the response.
            nonce: A 16 to 40 character long string with random unique data
            timestamp: Timestamp=1 requests timestamp and session
                       counter information in the response
            timeout: Number of seconds to wait for sync responses;
                     if absent, let the server decide
            sync_level: A value 0 to 100 indicating percentage of
                        syncing required by client, or strings "fast" or
                        "secure" to use server-configured values;
                        if absent, let the server decide

        Returns:
            A signed response with status=OK if the OTP is valid

        Raises:
            YKValError('BAD_OTP'): The OTP is invalid format
            YKValError('REPLAYED_OTP'): The OTP has already been seen by the service
            YKValError('MISSING_PARAMETER'): The request lacks a parameter
            YKValError('NO_SUCH_CLIENT'): The request id does not exist
            YKValError('OPERATION_NOT_ALLOWED'): The request id is not allowed to verify OTPs
            YKValError('BACKEND_ERROR'): Unexpected error in the server
            YKValError('NOT_ENOUGH_ANSWERS'): Server could not get requested number
                                              of syncs during before timeout
            YKValError('REPLAYED_REQUEST'): Server has seen the OTP/Nonce combination before


        Verify OTP process:
            1. sanitize input parameters
            2. decrypt OTP (YKKSM)
            3. compare old OTP counters with the given OTP counters and check for replay
            4. replicate new OTP counters to remote servers and check for replay on other servers
            5. check for phishing: OTP has to be used within a timeframe otherwise mark as expired
            6. prepare response: Sign the response with the right client key
        """

        ###################################
        # STEP 1: sanitize input parameters
        ###################################
        self.timeout = timeout if timeout else settings['SYNC_TIMEOUT']
        self.sync_level = sync_level if sync_level else settings['SYNC_LEVEL']
        server_nonce = generate_nonce()
        params = {
            'client_id': client_id,
            'otp': otp,
            'nonce': nonce if nonce else server_nonce,
            'timestamp': timestamp,
            'timeout': self.timeout,
            'sync_level': self.sync_level,
        }
        extra_params = {
            'otp': otp,
            'nonce': params['nonce']
        }
        # Check sanity of parameters
        self.check_parameters(params)

        #####################
        # STEP 2: decrypt OTP
        #####################
        otp_info = self.decode_otp(otp)

        #######################################
        # STEP 3: compare old OTP counters with
        #         the given OTP counters and
        #         check for replay
        #######################################
        # Get old parameters (counters) for the token
        local_params = self.db.get_local_params(otp[:-TOKEN_LEN])
        if not local_params['active']:
            logger.error('De-activated Yubikey: %(yk_publicname)s', local_params)
            raise YKValError('DISABLED_TOKEN')
        # Build the new parameters (counters) for the given OTP
        otp_params = self.build_otp_params(params, otp_info)
        # Validate OTP, check for replayed request or replayed OTP
        self.validate_otp(otp_params, local_params)

        #####################################
        # STEP 4: replicate new OTP counters
        #         to remote servers and check
        #         for replay on other servers
        #####################################
        sync_level_success_rate = self.replicate(otp_params, local_params, server_nonce)

        #######################################
        # STEP 5: check for phishing, OTP has
        #         to be used within a timeframe
        #         otherwise mark as expired
        #######################################
        self.phishing_test(otp_params, local_params)

        ##########################
        # STEP 6: Prepare response
        ##########################
        extra_params['sl'] = sync_level_success_rate
        if timestamp == 1:
            extra_params['timestamp'] = (otp_info['yk_high'] << 16) + otp_info['yk_low']
            extra_params['sessioncounter'] = otp_info['yk_counter']
            extra_params['sessionuse'] = otp_info['yk_use']
        response = {
            'status': 'OK',
            'time': datetime.utcnow().isoformat().replace('.', 'Z')[:-2]
        }
        response.update(extra_params)
        return response
